from warnings import warn
from ._venn import generate_petal_labels, draw_venn
from functools import partial

OLD_COLORS = [
    [.36, .75, .38, .50],
    [.35, .61, .83, .50],
    [.96, .93, .34, .60],
    [.95, .35, .38, .40],
    [.99, .46, .00, .30],
    [.32, .32, .75, .20]
]

def get_labels(data, fill=["number"]):
    warn((
        "`get_labels()` is retained for backwards compatibility; " +
        "use `generate_petal_labels()` or the higher level `venn()` instead"
    ))
    fmt = ""
    if "logic" in fill:
        fmt += "{logic}: "
    if "number" in fill:
        fmt += "{size} "
    if "percent" in fill:
        fmt += "({percentage:.1f}%)"
    return generate_petal_labels(data, fmt)

def vennx(labels, names, colors=OLD_COLORS, figsize=(9, 9), dpi=None, fontsize=14):
    n_sets = len(names)
    warn((
        "`venn{}()` is retained for backwards compatibility; ".format(n_sets) +
        "use `venn()` instead"
    ))
    if dpi is not None:
        warn("Option `dpi` is deprecated and has no effect")
    ax = draw_venn(
        petal_labels=labels, dataset_labels=names,
        colors=colors, figsize=figsize, fontsize=fontsize,
        legend_loc="best", ax=None
    )
    return ax.figure, ax

venn2 = partial(vennx, names=["AB"])
venn3 = partial(vennx, names=["ABC"])
venn4 = partial(vennx, names=["ABCD"])
venn5 = partial(vennx, names=["ABCDE"])
venn6 = partial(vennx, names=["ABCDEF"])



from matplotlib.pyplot import subplots
from matplotlib.patches import Ellipse, Polygon
from matplotlib.colors import to_rgba
from matplotlib.cm import ScalarMappable
from ._constants import SHAPE_COORDS, SHAPE_DIMS, SHAPE_ANGLES
from ._constants import PETAL_LABEL_COORDS, PSEUDOVENN_PETAL_COORDS
from math import pi, sin, cos
from functools import partial

def generate_colors(n_colors=6, cmap="viridis", alpha=.4):
    """Generate colors from matplotlib colormap; pass list to use exact colors"""
    if not isinstance(n_colors, int) or (n_colors < 2) or (n_colors > 6):
        raise ValueError("n_colors must be an integer between 2 and 6")
    if isinstance(cmap, list):
        colors = [to_rgba(color, alpha=alpha) for color in cmap]
    else:
        scalar_mappable = ScalarMappable(cmap=cmap)
        colors = scalar_mappable.to_rgba(range(n_colors), alpha=alpha).tolist()
    return colors[:n_colors]

def less_transparent_color(color, alpha_factor=2):
    """Bump up color's alpha"""
    new_alpha = (1 + to_rgba(color)[3]) / alpha_factor
    return to_rgba(color, alpha=new_alpha)

def draw_ellipse(ax, x, y, w, h, a, color):
    """Wrapper for drawing ellipse; called like `draw_ellipse(ax, *coords, *dims, angle, color)`"""
    ax.add_patch(
        Ellipse(
            xy=(x,y), width=w, height=h, angle=a,
            facecolor=color, edgecolor=less_transparent_color(color)
        )
    )

def draw_triangle(ax, x1, y1, x2, y2, x3, y3, _dim, _angle, color):
    """Wrapper for drawing triangle; called like `draw_triangle(ax, *coords, None, None, color)`"""
    ax.add_patch(
        Polygon(
            xy=[(x1, y1), (x2, y2), (x3, y3)], closed=True,
            facecolor=color, edgecolor=less_transparent_color(color)
        )
    )

def draw_text(ax, x, y, text, fontsize, color="black"):
    """Wrapper for drawing text"""
    ax.text(
        x, y, text, fontsize=fontsize, color=color,
        horizontalalignment="center", verticalalignment="center"
    )

def generate_logics(n_sets):
    """Generate intersection identifiers in binary (0010 etc)"""
    for i in range(1, 2**n_sets):
        yield bin(i)[2:].zfill(n_sets)

def generate_petal_labels(datasets, fmt="{size}"):
    """Generate petal descriptions for venn diagram based on set sizes"""
    datasets = list(datasets)
    n_sets = len(datasets)
    dataset_union = set.union(*datasets)
    universe_size = len(dataset_union)
    petal_labels = {}
    for logic in generate_logics(n_sets):
        included_sets = [
            datasets[i] for i in range(n_sets) if logic[i] == "1"
        ]
        excluded_sets = [
            datasets[i] for i in range(n_sets) if logic[i] == "0"
        ]
        petal_set = (
            (dataset_union & set.intersection(*included_sets)) -
            set.union(set(), *excluded_sets)
        )
        petal_labels[logic] = fmt.format(
            logic=logic, size=len(petal_set),
            percentage=(100*len(petal_set)/universe_size)
        )
    return petal_labels

def init_axes(ax, figsize):
    """Create axes if do not exist, set axes parameters"""
    if ax is None:
        _, ax = subplots(nrows=1, ncols=1, figsize=figsize)
    ax.set(
        aspect="equal", frame_on=False,
        xlim=(-.05, 1.05), ylim=(-.05, 1.05),
        xticks=[], yticks=[]
    )
    return ax

def get_n_sets(petal_labels, dataset_labels):
    """Infer number of sets, check consistency"""
    n_sets = len(dataset_labels)
    for logic in petal_labels.keys():
        if len(logic) != n_sets:
            raise ValueError("Inconsistent petal and dataset labels")
        if not (set(logic) <= {"0", "1"}):
            raise KeyError("Key not understood: " + logic)
    return n_sets

def draw_venn(*, petal_labels, dataset_labels, colors, figsize, fontsize, legend_loc, ax):
    """Draw true Venn diagram, annotate petals and dataset labels"""
    n_sets = get_n_sets(petal_labels, dataset_labels)
    if 2 <= n_sets < 6:
        draw_shape = draw_ellipse
    elif n_sets == 6:
        draw_shape = draw_triangle
    else:
        raise ValueError("Number of sets must be between 2 and 6")
    ax = init_axes(ax, figsize)
    shape_params = zip(
        SHAPE_COORDS[n_sets], SHAPE_DIMS[n_sets], SHAPE_ANGLES[n_sets], colors
    )
    for coords, dims, angle, color in shape_params:
        draw_shape(ax, *coords, *dims, angle, color)
    for logic, petal_label in petal_labels.items():
        # some petals could have been modified manually:
        if logic in PETAL_LABEL_COORDS[n_sets]:
            x, y = PETAL_LABEL_COORDS[n_sets][logic]
            draw_text(ax, x, y, petal_label, fontsize=fontsize)
    if legend_loc is not None:
        ax.legend(dataset_labels, loc=legend_loc, prop={"size": fontsize})
    return ax

def draw_pseudovenn6(*, petal_labels, dataset_labels, colors, figsize, fontsize, legend_loc, ax):
    """Draw intersection of 6 circles (does not include some combinations), annotate petals and dataset labels"""
    n_sets = get_n_sets(petal_labels, dataset_labels)
    if n_sets != 6:
        raise NotImplementedError("Pseudovenn implemented only for 6 sets")
    ax = init_axes(ax, figsize)
    for step, color in zip(range(6), colors):
        angle = (2 - step) * pi / 3
        x = .5 + .2 * cos(angle)
        y = .5 + .2 * sin(angle)
        draw_ellipse(ax, x, y, .6, .6, 0, color)
    for logic, petal_label in petal_labels.items():
        # not all theoretical intersections are shown, and petals could have been modified manually:
        if logic in PSEUDOVENN_PETAL_COORDS[6]:
            x, y = PSEUDOVENN_PETAL_COORDS[6][logic]
            draw_text(ax, x, y, petal_label, fontsize=fontsize)
    if legend_loc is not None:
        ax.legend(dataset_labels, loc=legend_loc, prop={"size": fontsize})
    return ax

def is_valid_dataset_dict(dataset_dict):
    """Validate passed data (must be dictionary of sets)"""
    if not (hasattr(dataset_dict, "keys") and hasattr(dataset_dict, "values")):
        return False
    for dataset in dataset_dict.values():
        if not isinstance(dataset, set):
            return False
    else:
        return True

def venn_dispatch(dataset_dict, draw_func, fmt="{size}", cmap="viridis", alpha=.4, figsize=(8, 8), fontsize=13, legend_loc="upper right", ax=None):
    """Check input, generate petal labels, draw venn or pseudovenn diagram"""
    if not is_valid_dataset_dict(dataset_dict):
        raise TypeError("Only dictionaries of sets are understood")
    n_sets = len(dataset_dict)
    return draw_func(
        petal_labels=generate_petal_labels(dataset_dict.values(), fmt),
        dataset_labels=dataset_dict.keys(),
        colors=generate_colors(n_colors=n_sets, cmap=cmap, alpha=alpha),
        figsize=figsize, fontsize=fontsize, legend_loc=legend_loc, ax=ax
    )

venn = partial(venn_dispatch, draw_func=draw_venn)
pseudovenn = partial(venn_dispatch, draw_func=draw_pseudovenn6)



from ._venn import venn, pseudovenn
from ._venn import generate_petal_labels, generate_colors
from ._venn import draw_venn, draw_pseudovenn6
from ._backwards_compatibility import get_labels
from ._backwards_compatibility import venn2, venn3, venn4, venn5, venn6



import setuptools
from argparse import Namespace
from subprocess import Popen, PIPE
from re import sub

meta = Namespace(
    __name__ = "venn",
    __version__ = "0.1.2",
    __author__ = "Kirill Grigorev",
    __git_id__ = "LankyCyril",
    __license__ = "GPLv3",
)

def get_remote():
    cmd = ["git", "remote", "-v", "show", "origin"]
    git = Popen(cmd, stdout=PIPE, universal_newlines=True)
    stdout = git.communicate()[0]
    try:
        fetch = stdout.split("\n")[1]
    except IndexError:
        return ""
    raw_remote = fetch.split("Fetch URL: ")[1]
    pattern = "(://){}@".format(meta.__git_id__)
    remote = sub(pattern, r"\1", raw_remote)
    return remote

def get_readme(filename):
    with open(filename) as readme_handle:
        return readme_handle.read()

if __name__ == "__main__":
    setuptools.setup(
        name = meta.__name__,
        version = meta.__version__,
        packages = [meta.__name__],
        url = get_remote(),
        author = meta.__author__,
        license = meta.__license__,
        zip_safe = True,
        description = "Venn diagrams for 2, 3, 4, 5, 6 sets",
        long_description = get_readme("README.md"),
        long_description_content_type = "text/markdown",
        classifiers = [
            "Development Status :: 3 - Alpha",
            "Programming Language :: Python :: 3",
            "Environment :: Console",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering :: Visualization"
        ],
        install_requires = ["matplotlib"]
    )
